# Generated by Django 2.2.5 on 2019-12-20 11:44

import django.contrib.postgres.fields.jsonb
from django.db import migrations


class Migration(migrations.Migration):
    deploy_phase = "pre_deploy"

    dependencies = [("mds", "0001_squashed_0059_post_policy_config_encoder")]

    operations = [
        migrations.AddField(
            model_name="compliance",
            name="extra",
            field=django.contrib.postgres.fields.jsonb.JSONField(blank=True, null=True),
        )
    ]
